#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This file is part of PLCOpenEditor, a library implementing an IEC 61131-3 editor
#based on the plcopen standard. 
#
#Copyright (C) 2007: Edouard TISSERANT and Laurent BESSARD
#
#See COPYING file for copyrights details.
#
#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.
#
#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#General Public License for more details.
#
#You should have received a copy of the GNU General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from xml.dom import minidom
from types import StringType, UnicodeType
import cPickle
import os,sys,re
from time import localtime
from datetime import *

from plcopen import plcopen
from plcopen.structures import *
from graphics.GraphicCommons import *
from PLCGenerator import *

duration_model = re.compile("(?:([0-9]{1,2})h)?(?:([0-9]{1,2})m(?!s))?(?:([0-9]{1,2})s)?(?:([0-9]{1,3}(?:\.[0-9]*)?)ms)?")

ITEMS_EDITABLE = [ITEM_PROJECT,
                    ITEM_POU,
                    ITEM_VARIABLE,
                    ITEM_TRANSITION,
                    ITEM_ACTION,
                    ITEM_CONFIGURATION,
                    ITEM_RESOURCE,
                    ITEM_DATATYPE] = range(8)

ITEMS_UNEDITABLE=[ITEM_DATATYPES,
                  ITEM_FUNCTION,
                  ITEM_FUNCTIONBLOCK,
                  ITEM_PROGRAM,
                  ITEM_TRANSITIONS,
                  ITEM_ACTIONS,
                  ITEM_CONFIGURATIONS,
                  ITEM_RESOURCES,
                  ITEM_PROPERTIES,
                  ]=range(9,18)

ScriptDirectory = os.path.split(os.path.realpath(__file__))[0]

#-------------------------------------------------------------------------------
#                         Undo Buffer for PLCOpenEditor
#-------------------------------------------------------------------------------

# Length of the buffer
UNDO_BUFFER_LENGTH = 20

"""
Class implementing a buffer of changes made on the current editing model
"""
class UndoBuffer:

    # Constructor initialising buffer
    def __init__(self, currentstate, issaved = False):
        self.Buffer = []
        self.CurrentIndex = -1
        self.MinIndex = -1
        self.MaxIndex = -1
        # if current state is defined
        if currentstate:
            self.CurrentIndex = 0
            self.MinIndex = 0
            self.MaxIndex = 0
        # Initialising buffer with currentstate at the first place
        for i in xrange(UNDO_BUFFER_LENGTH):
            if i == 0:
                self.Buffer.append(currentstate)
            else:
                self.Buffer.append(None)
        # Initialising index of state saved
        if issaved:
            self.LastSave = 0
        else:
            self.LastSave = -1
    
    # Add a new state in buffer
    def Buffering(self, currentstate):
        self.CurrentIndex = (self.CurrentIndex + 1) % UNDO_BUFFER_LENGTH
        self.Buffer[self.CurrentIndex] = currentstate
        # Actualising buffer limits
        self.MaxIndex = self.CurrentIndex
        if self.MinIndex == self.CurrentIndex:
            # If the removed state was the state saved, there is no state saved in the buffer
            if self.LastSave == self.MinIndex:
                self.LastSave = -1
            self.MinIndex = (self.MinIndex + 1) % UNDO_BUFFER_LENGTH
        self.MinIndex = max(self.MinIndex, 0)
    
    # Return current state of buffer
    def Current(self):
        return self.Buffer[self.CurrentIndex]
    
    # Change current state to previous in buffer and return new current state
    def Previous(self):
        if self.CurrentIndex != self.MinIndex:
            self.CurrentIndex = (self.CurrentIndex - 1) % UNDO_BUFFER_LENGTH
            return self.Buffer[self.CurrentIndex]
        return None
    
    # Change current state to next in buffer and return new current state
    def Next(self):
        if self.CurrentIndex != self.MaxIndex:
            self.CurrentIndex = (self.CurrentIndex + 1) % UNDO_BUFFER_LENGTH
            return self.Buffer[self.CurrentIndex]
        return None
    
    # Return True if current state is the first in buffer
    def IsFirst(self):
        return self.CurrentIndex == self.MinIndex
    
    # Return True if current state is the last in buffer
    def IsLast(self):
        return self.CurrentIndex == self.MaxIndex

    # Note that current state is saved
    def CurrentSaved(self):
        self.LastSave = self.CurrentIndex
        
    # Return True if current state is saved
    def IsCurrentSaved(self):
        return self.LastSave == self.CurrentIndex


#-------------------------------------------------------------------------------
#                           Controler for PLCOpenEditor
#-------------------------------------------------------------------------------

"""
Class which controls the operations made on the plcopen model and answers to view requests
"""
class PLCControler:
    
    # Create a new PLCControler
    def __init__(self):
        self.LastNewIndex = 0
        self.Reset()
    
    # Reset PLCControler internal variables
    def Reset(self):
        self.Project = None
        self.ProjectBuffer = None
        self.Buffering = False
        self.FilePath = ""
        self.FileName = ""
        self.ProgramChunks = []
        self.PluginTypes = []
        self.ProgramFilePath = ""
        
    def GetQualifierTypes(self):
        return plcopen.QualifierList

#-------------------------------------------------------------------------------
#                         Project management functions
#-------------------------------------------------------------------------------

    # Return if a project is opened
    def HasOpenedProject(self):
        return self.Project != None

    # Create a new project by replacing the current one
    def CreateNewProject(self, properties):
        # Create the project
        self.Project = plcopen.project()
        properties["creationDateTime"] = datetime(*localtime()[:6])
        self.Project.setfileHeader(properties)
        self.Project.setcontentHeader(properties)
        self.SetFilePath("")
        # Initialize the project buffer
        self.ProjectBuffer = UndoBuffer(self.Copy(self.Project), False)
        self.Buffering = False
    
    # Return project data type names
    def GetProjectDataTypeNames(self):
        if self.Project:
            return [datatype.getname() for datatype in self.Project.getdataTypes()]
        return []
    
    # Return project pou names
    def GetProjectPouNames(self):
        if self.Project:
            return [pou.getname() for pou in self.Project.getpous()]
        return []
    
    # Return project pou names
    def GetProjectConfigNames(self):
        if self.Project:
            return [config.getName() for config in self.Project.getconfigurations()]
        return []
    
    # Return project pou variables
    def GetProjectPouVariables(self, pou_name=None):
        variables = []
        for pou in self.Project.getpous():
            if not pou_name or pou_name == pou.getname():
                variables.extend([var["Name"] for var in self.GetPouInterfaceVars(pou)])
                for transition in pou.gettransitionList():
                    variables.append(transition.getname())
                for action in pou.getactionList():
                    variables.append(action.getname())
        return variables
    
    # Return if project is saved
    def ProjectIsSaved(self):
        return self.ProjectBuffer.IsCurrentSaved()
    
    # Return file path if project is an open file
    def GetFilePath(self):
        return self.FilePath
    
    # Return file path if project is an open file
    def GetProgramFilePath(self):
        return self.ProgramFilePath
    
    # Return file name and point out if file is up to date
    def GetFilename(self):
        if self.ProjectBuffer.IsCurrentSaved():
            return self.FileName
        else:
            return "~%s~"%self.FileName
    
    # Change file path and save file name or create a default one if file path not defined
    def SetFilePath(self, filepath):
        self.FilePath = filepath
        if filepath == "":
            self.LastNewIndex += 1
            self.FileName = "Unnamed%d"%self.LastNewIndex
        else:
            self.FileName = os.path.splitext(os.path.basename(filepath))[0]
    
    # Change project properties
    def SetProjectProperties(self, name = None, properties = None):
        if name != None:
            self.Project.setname(name)
        if properties != None:
            self.Project.setfileHeader(properties)
            self.Project.setcontentHeader(properties)
        if name != None or properties != None:
            self.BufferProject()
            
    # Return project properties
    def GetProjectProperties(self):
        properties = self.Project.getfileHeader()
        properties.update(self.Project.getcontentHeader())
        return properties
    
    # Return project informations
    def GetProjectInfos(self):
        if self.Project:
            infos = {"name": self.Project.getname(), "type": ITEM_PROJECT, "tagname": ""}
            datatypes = {"name": "Data Types", "type": ITEM_DATATYPES, "tagname": "", "values":[]}
            for datatype in self.Project.getdataTypes():
                datatypes["values"].append({"name": datatype.getname(), "type": ITEM_DATATYPE, 
                    "tagname": self.ComputeDataTypeName(datatype.getname()), "values": []})
            pou_types = {"function": {"name": "Functions", "type": ITEM_FUNCTION, "tagname": "", "values":[]},
                         "functionBlock": {"name": "Function Blocks", "type": ITEM_FUNCTIONBLOCK, "tagname": "", "values":[]},
                         "program": {"name": "Programs", "type": ITEM_PROGRAM, "tagname": "", "values":[]}}
            for pou in self.Project.getpous():
                pou_type = pou.getpouType()
                pou_infos = {"name": pou.getname(), "type": ITEM_POU,
                             "tagname": self.ComputePouName(pou.getname())}
                pou_values = []
                if pou.getbodyType() == "SFC":
                    transitions = []
                    for transition in pou.gettransitionList():
                        transitions.append({"name": transition.getname(), "type": ITEM_TRANSITION, 
                            "tagname": self.ComputePouTransitionName(pou.getname(), transition.getname()), 
                            "values": []})
                    pou_values.append({"name": "Transitions", "type": ITEM_TRANSITIONS, 
                                       "tagname": "", "values": transitions})
                    actions = []
                    for action in pou.getactionList():
                        actions.append({"name": action.getname(), "type": ITEM_ACTION, 
                            "tagname": self.ComputePouActionName(pou.getname(), action.getname()), 
                            "values": []})
                    pou_values.append({"name": "Actions", "type": ITEM_ACTIONS, 
                                       "tagname": "", "values": actions})
                if pou_type in pou_types:
                    pou_infos["values"] = pou_values
                    pou_types[pou_type]["values"].append(pou_infos)
            configurations = {"name": "Configurations", "type": ITEM_CONFIGURATIONS, 
                              "tagname": "", "values": []}
            for config in self.Project.getconfigurations():
                config_name = config.getname()
                config_infos = {"name": config_name, "type": ITEM_CONFIGURATION, 
                    "tagname": self.ComputeConfigurationName(config.getname()), 
                    "values": []}
                resources = {"name": "Resources", "type": ITEM_RESOURCES, "tagname": "", "values": []}
                for resource in config.getresource():
                    resource_name = resource.getname()
                    resource_infos = {"name": resource_name, "type": ITEM_RESOURCE, 
                        "tagname": self.ComputeConfigurationResourceName(config.getname(), resource.getname()), 
                        "values": []}
                    resources["values"].append(resource_infos)
                config_infos["values"] = [resources]
                configurations["values"].append(config_infos)
            infos["values"] = [{"name": "Properties", "type": ITEM_PROPERTIES, "tagname": "", "values": []},
                               datatypes, pou_types["function"], pou_types["functionBlock"], 
                               pou_types["program"], configurations]
            return infos
        return None

    # Return project topology informations
    def GetProjectTopology(self):
        if self.Project:
            infos = {"name": self.Project.getname(), "type": ITEM_PROJECT, "values" : []}
            for config in self.Project.getconfigurations():
                config_infos = {"name" : config.getname(), "type": ITEM_CONFIGURATION, "values" : []}
                for resource in config.getresource():
                    resource_infos = {"name" : resource.getname(), "type": ITEM_RESOURCE, "values": []}
                    for task in resource.gettask():
                        for pou in task.getpouInstance():
                            instance_infos = self.GetPouTopology(pou.getname(), pou.gettype())
                            if instance_infos is not None:
                                resource_infos["values"].append(instance_infos)
                    for pou in resource.getpouInstance():
                        instance_infos = self.GetPouTopology(pou.getname(), pou.gettype())
                        if instance_infos is not None:
                            resource_infos["values"].append(instance_infos)
                    for varlist in resource.getglobalVars():
                        for variable in varlist.getvariable():
                            vartype_content = variable.gettype().getcontent()
                            if vartype_content["name"] == "derived":
                                var_infos = self.GetPouTopology(variable.getname(), vartype_content["value"].getname())
                                if var_infos is not None:
                                    resource_infos["values"].append(var_infos)
                            elif vartype_content["name"] in ["string", "wstring"]:
                                resource_infos["values"].append({"name" : "%s(%s)"%(variable.getname(), vartype_content["name"].upper()), 
                                                                 "type" : ITEM_VARIABLE, "values" : []})
                            else:
                                resource_infos["values"].append({"name" : "%s(%s)"%(variable.getname(), vartype_content["name"]), 
                                                                 "type" : ITEM_VARIABLE, "values" : []})
                    config_infos["values"].append(resource_infos)
                for varlist in config.getglobalVars():
                    for variable in varlist.getvariable():
                        vartype_content = variable.gettype().getcontent()
                        if vartype_content["name"] == "derived":
                            var_infos = self.GetPouTopology(variable.getname(), vartype_content["value"].getname())
                            if var_infos is not None:
                                config_infos["values"].append(var_infos)
                        elif vartype_content["name"] in ["string", "wstring"]:
                            config_infos["values"].append({"name" : "%s(%s)"%(variable.getname(), vartype_content["name"].upper()), 
                                                           "type" : ITEM_VARIABLE, "values" : []})
                        else:
                            config_infos["values"].append({"name" : "%s(%s)"%(variable.getname(), vartype_content["name"]), 
                                                           "type" : ITEM_VARIABLE, "values" : []})
                infos["values"].append(config_infos)
            return infos
        return None
    
    # Return pou topology informations
    def GetPouTopology(self, name, type):
        if self.Project:
            pou = self.Project.getpou(type)
            if pou is not None:
                pou_infos = {"name" : "%s(%s)"%(name, type), "type" : ITEM_POU, "values" : []}
                if pou.interface:
                    # Extract variables from every varLists
                    for type, varlist in pou.getvars():
                        for variable in varlist.getvariable():
                            vartype_content = variable.gettype().getcontent()
                            if vartype_content["name"] == "derived":
                                var_infos = self.GetPouTopology(variable.getname(), vartype_content["value"].getname())
                                if var_infos is not None:
                                    pou_infos["values"].append(var_infos)
                            elif vartype_content["name"] in ["string", "wstring"]:
                                pou_infos["values"].append({"name" : "%s(%s)"%(variable.getname(), vartype_content["name"].upper()), 
                                                               "type" : ITEM_VARIABLE, "values" : []})
                            else:
                                pou_infos["values"].append({"name" : "%s(%s)"%(variable.getname(), vartype_content["name"]), 
                                                               "type" : ITEM_VARIABLE, "values" : []})
                return pou_infos
            block_infos = self.GetBlockType(type)
            if block_infos is not None and block_infos["type"] != "function":
                pou_infos = {"name" : "%s(%s)"%(name, type), "type" : ITEM_POU, "values" : []}
                for varname, vartype, varmodifier in block_infos["inputs"]:
                    pou_infos["values"].append({"name" : "%s(%s)"%(varname, vartype), "type" : ITEM_VARIABLE, "values" : []})
                for varname, vartype, varmodifier in block_infos["outputs"]:
                    pou_infos["values"].append({"name" : "%s(%s)"%(varname, vartype), "type" : ITEM_VARIABLE, "values" : []})
                return pou_infos
            
            if type in TypeHierarchy:
                return {"name" : "%s(%s)"%(name, type), "type" : ITEM_VARIABLE, "values" : []}
                
        return None

    # Return if data type given by name is used by another data type or pou
    def DataTypeIsUsed(self, name):
        return self.Project.ElementIsUsed(name) or self.Project.DataTypeIsDerived(name)

    # Return if pou given by name is used by another pou
    def PouIsUsed(self, name):
        return self.Project.ElementIsUsed(name)

    # Return if pou given by name is directly or undirectly used by the reference pou
    def PouIsUsedBy(self, name, reference):
        return self.Project.ElementIsUsedBy(name, reference)

    def GenerateProgram(self, filepath):
        if self.Project:
            try:
                self.ProgramChunks = GenerateCurrentProgram(self, self.Project)
                program_text = "".join([item[0] for item in self.ProgramChunks])
                programfile = open(filepath, "w")
                programfile.write(program_text)
                programfile.close()
                self.ProgramFilePath = filepath
                return None
            except PLCGenException, e:
                return e.message
        return "No project opened"

    def GetChunkInfos(self, from_location, to_location):
        row = col = 1
        infos = []
        for chunk, chunk_infos in self.ProgramChunks:
            lines = chunk.split("\n")
            if len(lines) > 1:
                next_row = row + len(lines) - 1
                next_col = len(lines[-1]) + 1
            else:
                next_col = col + len(chunk)
            if (next_row > from_location[0] or next_row == from_location[0] and next_col >= from_location[1]) and len(chunk_infos) > 0:
                infos.append((chunk_infos, (row, col)))
            if next_row == to_location[0] and next_col > to_location[1] or next_row > to_location[0]:
                return infos
            row, col = next_row, next_col
        return infos
        
#-------------------------------------------------------------------------------
#                        Project Pous management functions
#-------------------------------------------------------------------------------
    
    # Add a Data Type to Project
    def ProjectAddDataType(self, datatype_name):
        # Add the pou to project
        self.Project.appenddataType(datatype_name)
        self.BufferProject()
    
    # Remove a Data Type from project
    def ProjectRemoveDataType(self, datatype_name):
        self.Project.removedataType(datatype_name)
        self.BufferProject()
    
    # Add a Pou to Project
    def ProjectAddPou(self, pou_name, pou_type, body_type):
        # Add the pou to project
        self.Project.appendpou(pou_name, pou_type, body_type)
        if pou_type == "function":
            self.SetPouInterfaceReturnType(pou_name, "BOOL")
        self.BufferProject()
    
    # Remove a Pou from project
    def ProjectRemovePou(self, pou_name):
        self.Project.removepou(pou_name)
        self.BufferProject()
    
    # Add a configuration to Project
    def ProjectAddConfiguration(self, config_name):
        self.Project.addconfiguration(config_name)
        self.BufferProject()
    
    # Remove a configuration from project
    def ProjectRemoveConfiguration(self, config_name):
        self.Project.removeconfiguration(config_name)
        self.BufferProject()
    
    # Add a resource to a configuration of the Project
    def ProjectAddConfigurationResource(self, config_name, resource_name):
        self.Project.addconfigurationResource(config_name, resource_name)
        self.BufferProject()
    
    # Remove a resource from a configuration of the project
    def ProjectRemoveConfigurationResource(self, config_name, resource_name):
        self.Project.removeconfigurationResource(config_name, resource_name)
        self.BufferProject()
    
    # Add a Transition to a Project Pou
    def ProjectAddPouTransition(self, pou_name, transition_name, transition_type):
        pou = self.Project.getpou(pou_name)
        pou.addtransition(transition_name, transition_type)
        self.BufferProject()
    
    # Remove a Transition from a Project Pou
    def ProjectRemovePouTransition(self, pou_name, transition_name):
        pou = self.Project.getpou(pou_name)
        pou.removetransition(transition_name)
        self.BufferProject()
    
    # Add an Action to a Project Pou
    def ProjectAddPouAction(self, pou_name, action_name, action_type):
        pou = self.Project.getpou(pou_name)
        pou.addaction(action_name, action_type)
        self.BufferProject()
    
    # Remove an Action from a Project Pou
    def ProjectRemovePouAction(self, pou_name, action_name):
        # Search if the pou removed is currently opened
        for i, element in enumerate(self.ElementsOpened):
            words = element.split("::")
            if words[0] == "A" and words[1] == pou_name and words[2] == action_name:
                self.RemoveElementEditing(i)
        pou = self.Project.getpou(pou_name)
        pou.removeaction(action_name)
        self.BufferProject()
    
    # Change the name of a pou
    def ChangeDataTypeName(self, old_name, new_name):
        # Found the pou corresponding to old name and change its name to new name
        datatype = self.Project.getdataType(old_name)
        datatype.setname(new_name)
        self.Project.updateElementName(old_name, new_name)
        self.BufferProject()
    
    # Change the name of a pou
    def ChangePouName(self, old_name, new_name):
        # Found the pou corresponding to old name and change its name to new name
        pou = self.Project.getpou(old_name)
        pou.setname(new_name)
        self.Project.updateElementName(old_name, new_name)
        self.Project.RefreshElementUsingTree()
        self.Project.RefreshCustomBlockTypes()
        self.BufferProject()
    
    # Change the name of a pou transition
    def ChangePouTransitionName(self, pou_name, old_name, new_name):
        # Found the pou transition corresponding to old name and change its name to new name
        pou = self.Project.getpou(pou_name)
        transition = pou.gettransition(old_name)
        transition.setname(new_name)
        pou.updateElementName(old_name, new_name)
        self.BufferProject()
    
    # Change the name of a pou action
    def ChangePouActionName(self, pou_name, old_name, new_name):
        # Found the pou action corresponding to old name and change its name to new name
        pou = self.Project.getpou(pou_name)
        action = pou.getaction(old_name)
        action.setname(new_name)
        pou.updateElementName(old_name, new_name)
        self.BufferProject()
    
    # Change the name of a pou variable
    def ChangePouVariableName(self, pou_name, old_name, new_name):
        # Found the pou action corresponding to old name and change its name to new name
        pou = self.Project.getpou(pou_name)
        for type, varlist in pou.getvars():
            for var in varlist.getvariable():
                if var.getname() == old_name:
                    var.setname(new_name)
        self.Project.RefreshCustomBlockTypes()
        self.BufferProject()
        
    # Change the name of a configuration
    def ChangeConfigurationName(self, old_name, new_name):
        # Found the configuration corresponding to old name and change its name to new name
        configuration = self.Project.getconfiguration(old_name)
        configuration.setname(new_name)
        self.BufferProject()
    
    # Change the name of a configuration resource
    def ChangeConfigurationResourceName(self, config_name, old_name, new_name):
        # Found the resource corresponding to old name and change its name to new name
        resource = self.Project.getconfigurationResource(config_name)
        resource.setName(new_name)
        self.BufferProject()
    
    # Return the type of the pou given by its name
    def GetPouType(self, name):
        # Found the pou correponding to name and return its type
        pou = self.Project.getpou(name)
        return pou.getpouType()
    
    # Return pous with SFC language
    def GetSFCPous(self):
        list = []
        if self.Project:
            for pou in self.Project.getpous():
                if pou.getBodyType() == "SFC":
                    list.append(pou.getname())
        return list
    
    # Return the body language of the pou given by its name
    def GetPouBodyType(self, name):
        # Found the pou correponding to name and return its body language
        pou = self.Project.getpou(name)
        return pou.getbodyType()
    
    # Return the actions of a pou
    def GetPouTransitions(self, pou_name):
        transitions = []
        pou = self.Project.getpou(pou_name)
        if pou.getbodyType() == "SFC":
            for transition in pou.gettransitionList():
                transitions.append(transition.getname())
        return transitions
    
    # Return the body language of the transition given by its name
    def GetTransitionBodyType(self, pou_name, pou_transition):
        # Found the pou correponding to name and return its body language
        pou = self.Project.getpou(pou_name)
        transition = pou.gettransition(pou_transition)
        return transition.getbodyType()
    
    # Return the actions of a pou
    def GetPouActions(self, pou_name):
        actions = []
        pou = self.Project.getpou(pou_name)
        if pou.getbodyType() == "SFC":
            for action in pou.getactionList():
                actions.append(action.getname())
        return actions
    
    # Return the body language of the pou given by its name
    def GetActionBodyType(self, pou_name, pou_action):
        # Found the pou correponding to name and return its body language
        pou = self.Project.getpou(pou_name)
        action = pou.getaction(pou_action)
        return action.getbodyType()
    
    # Extract varlists from a list of vars
    def ExtractVarLists(self, vars):
        varlist_list = []
        current_varlist = None
        current_type = None
        for var in vars:
            if current_type != (var["Class"], var["Retain"], var["Constant"], var["Location"] in ["", None]):
                current_type = (var["Class"], var["Retain"], var["Constant"], var["Location"] in ["", None])
                if var["Class"] == "Local":
                    current_varlist = plcopen.interface_localVars()
                elif var["Class"] == "Global":
                    current_varlist = plcopen.interface_globalVars()
                elif var["Class"] == "External":
                    current_varlist = plcopen.interface_externalVars()
                elif var["Class"] == "Temp":
                    current_varlist = plcopen.interface_tempVars()
                elif var["Class"] == "Input":
                    current_varlist = plcopen.interface_inputVars()
                elif var["Class"] == "Output":
                    current_varlist = plcopen.interface_outputVars()
                elif var["Class"] == "InOut":
                    current_varlist = plcopen.interface_inOutVars()
                else:
                    current_varlist = plcopen.varList()
                varlist_list.append((var["Class"], current_varlist))
                if var["Retain"] == "Yes":
                    current_varlist.setretain(True)
                if var["Constant"] == "Yes":
                    current_varlist.setconstant(True)
            # Create variable and change its properties
            tempvar = plcopen.varListPlain_variable()
            tempvar.setname(var["Name"])
            var_type = plcopen.dataType()
            if var["Type"] in self.GetBaseTypes():
                if var["Type"] == "STRING":
                    var_type.setcontent({"name" : "string", "value" : plcopen.elementaryTypes_string()})
                elif var["Type"] == "WSTRING":
                    var_type.setcontent({"name" : "wstring", "value" : plcopen.elementaryTypes_wstring()})
                else:
                    var_type.setcontent({"name" : var["Type"], "value" : None})
            else:
                derived_type = plcopen.derivedTypes_derived()
                derived_type.setname(var["Type"])
                var_type.setcontent({"name" : "derived", "value" : derived_type})
            tempvar.settype(var_type)
            if var["Initial Value"] != "":
                value = plcopen.value()
                value.setvalue(var["Initial Value"])
                tempvar.setinitialValue(value)
            if var["Location"] != "":
                tempvar.setaddress(var["Location"])
            else:
                tempvar.setaddress(None)
            # Add variable to varList
            current_varlist.appendvariable(tempvar)
        return varlist_list

    # Replace the configuration globalvars by those given
    def SetConfigurationGlobalVars(self, name, vars):
        # Found the configuration corresponding to name
        configuration = self.Project.getconfiguration(name)
        if configuration:
            # Set configuration global vars
            configuration.setglobalVars([])
            for vartype, varlist in self.ExtractVarLists(vars):
                configuration.globalVars.append(varlist)
    
    # Return the configuration globalvars
    def GetConfigurationGlobalVars(self, name):
        vars = []
        # Found the configuration corresponding to name
        configuration = self.Project.getconfiguration(name)
        if configuration:
            # Extract variables from every varLists
            for varlist in configuration.getglobalVars():
                for var in varlist.getvariable():
                    tempvar = {"Name" : var.getname(), "Class" : "Global"}
                    vartype_content = var.gettype().getcontent()
                    if vartype_content["name"] == "derived":
                        tempvar["Type"] = vartype_content["value"].getname()
                    elif vartype_content["name"] in ["string", "wstring"]:
                        tempvar["Type"] = vartype_content["name"].upper()
                    else:
                        tempvar["Type"] = vartype_content["name"]
                    tempvar["Edit"] = True
                    initial = var.getinitialValue()
                    if initial:
                        tempvar["Initial Value"] = initial.getvalue()
                    else:
                        tempvar["Initial Value"] = ""
                    address = var.getaddress()
                    if address:
                        tempvar["Location"] = address
                    else:
                        tempvar["Location"] = ""
                    if varlist.getretain():
                        tempvar["Retain"] = "Yes"
                    else:
                        tempvar["Retain"] = "No"
                    if varlist.getconstant():
                        tempvar["Constant"] = "Yes"
                    else:
                        tempvar["Constant"] = "No"
                    vars.append(tempvar)
        return vars

    # Replace the resource globalvars by those given
    def SetConfigurationResourceGlobalVars(self, config_name, name, vars):
        # Found the resource corresponding to name
        resource = self.Project.getconfigurationResource(config_name, name)
        # Set resource global vars
        if resource:
            resource.setglobalVars([])
            for vartype, varlist in self.ExtractVarLists(vars):
                resource.globalVars.append(varlist)
    
    # Return the resource globalvars
    def GetConfigurationResourceGlobalVars(self, config_name, name):
        vars = []
        # Found the resource corresponding to name
        resource = self.Project.getconfigurationResource(config_name, name)
        if resource:
            # Extract variables from every varLists
            for varlist in resource.getglobalVars():
                for var in varlist.getvariable():
                    tempvar = {"Name" : var.getname(), "Class" : "Global"}
                    vartype_content = var.gettype().getcontent()
                    if vartype_content["name"] == "derived":
                        tempvar["Type"] = vartype_content["value"].getname()
                    elif vartype_content["name"] in ["string", "wstring"]:
                        tempvar["Type"] = vartype_content["name"].upper()
                    else:
                        tempvar["Type"] = vartype_content["name"]
                    tempvar["Edit"] = True
                    initial = var.getinitialValue()
                    if initial:
                        tempvar["Initial Value"] = initial.getvalue()
                    else:
                        tempvar["Initial Value"] = ""
                    address = var.getaddress()
                    if address:
                        tempvar["Location"] = address
                    else:
                        tempvar["Location"] = ""
                    if varlist.getretain():
                        tempvar["Retain"] = "Yes"
                    else:
                        tempvar["Retain"] = "No"
                    if varlist.getconstant():
                        tempvar["Constant"] = "Yes"
                    else:
                        tempvar["Constant"] = "No"
                    vars.append(tempvar)
        return vars
    
    # Return the interface of the pou given by its name
    def GetPouInterfaceVarsByName(self, name):
        # Found the pou correponding to name and return the interface
        return self.GetPouInterfaceVars(self.Project.getpou(name))
    
    # Return the interface for the given pou
    def GetPouInterfaceVars(self, pou):
        vars = []
        # Verify that the pou has an interface
        if pou.interface:
            # Extract variables from every varLists
            for type, varlist in pou.getvars():
                for var in varlist.getvariable():
                    tempvar = {"Name" : var.getname(), "Class" : type}
                    vartype_content = var.gettype().getcontent()
                    if vartype_content["name"] == "derived":
                        tempvar["Type"] = vartype_content["value"].getname()
                        tempvar["Edit"] = not pou.hasblock(tempvar["Name"])
                    else:
                        if vartype_content["name"] in ["string", "wstring"]:
                            tempvar["Type"] = vartype_content["name"].upper()
                        else:
                            tempvar["Type"] = vartype_content["name"]
                        tempvar["Edit"] = True
                    initial = var.getinitialValue()
                    if initial:
                        tempvar["Initial Value"] = initial.getvalue()
                    else:
                        tempvar["Initial Value"] = ""
                    address = var.getaddress()
                    if address:
                        tempvar["Location"] = address
                    else:
                        tempvar["Location"] = ""
                    if varlist.getretain():
                        tempvar["Retain"] = "Yes"
                    else:
                        tempvar["Retain"] = "No"
                    if varlist.getconstant():
                        tempvar["Constant"] = "Yes"
                    else:
                        tempvar["Constant"] = "No"
                    vars.append(tempvar)
        return vars

    # Replace the Pou interface by the one given
    def SetPouInterfaceVars(self, name, vars):
        # Found the pou corresponding to name and add interface if there isn't one yet
        pou = self.Project.getpou(name)
        if not pou.interface:
            pou.interface = plcopen.pou_interface()
        # Set Pou interface
        pou.setvars(self.ExtractVarLists(vars))
        self.Project.RefreshElementUsingTree()
        self.Project.RefreshCustomBlockTypes()
    
    # Replace the return type of the pou given by its name (only for functions)
    def SetPouInterfaceReturnType(self, name, type):
        pou = self.Project.getpou(name)
        if not pou.interface:
            pou.interface = plcopen.pou_interface()
        # If there isn't any return type yet, add it
        return_type = pou.interface.getreturnType()
        if not return_type:
            return_type = plcopen.dataType()
            pou.interface.setreturnType(return_type)
        # Change return type
        if type in self.GetBaseTypes():
            if type == "STRING":
                return_type.setcontent({"name" : "string", "value" : plcopen.elementaryTypes_string()})
            elif type == "WSTRING":
                return_type.setcontent({"name" : "wstring", "value" : plcopen.elementaryTypes_wstring()})
            else:
                return_type.setcontent({"name" : type, "value" : None})
        else:
            derived_type = plcopen.derivedTypes_derived()
            derived_type.setname(type)
            return_type.setcontent({"name" : "derived", "value" : derived_type})
        self.Project.RefreshElementUsingTree()
        self.Project.RefreshCustomBlockTypes()
    
    def UpdateProjectUsedPous(self, old_name, new_name):
        if self.Project:
            self.Project.updateElementName(old_name, new_name)
    
    def UpdateEditedElementUsedVariable(self, tagname, old_name, new_name):
        pou = self.GetEditedElement(tagname)
        if pou:
            pou.updateElementName(old_name, new_name)
    
    # Return the return type of the pou given by its name
    def GetPouInterfaceReturnTypeByName(self, name):
        # Found the pou correponding to name and return the return type
        return self.GetPouInterfaceReturnType(self.Project.getpou(name))
    
    # Return the return type of the given pou
    def GetPouInterfaceReturnType(self, pou):
        # Verify that the pou has an interface
        if pou.interface:
            # Return the return type if there is one
            return_type = pou.interface.getreturnType()
            if return_type:
                returntype_content = return_type.getcontent()
                if returntype_content["name"] == "derived":
                    return returntype_content["value"].getname()
                elif returntype_content["name"] in ["string", "wstring"]:
                    return returntype_content["name"].upper()
                else:
                    return returntype_content["name"]
        return None

    # Function that add a new plugin to the plugin list
    def AddPluginBlockList(self, blocklist):
        self.PluginTypes.extend(blocklist)
        
    # Function that clear the plugin list
    def ClearPluginTypes(self):
        for i in xrange(len(self.PluginTypes)):
            self.PluginTypes.pop(0)

    # Function that returns the block definition associated to the block type given
    def GetBlockType(self, type, inputs = None):
        for category in BlockTypes + self.PluginTypes:
            for blocktype in category["list"]:
                if inputs:
                    block_inputs = tuple([var_type for name, var_type, modifier in blocktype["inputs"]])
                    same_inputs = inputs == block_inputs
                else:
                    same_inputs = True
                if blocktype["name"] == type and same_inputs:
                    return blocktype
        if self.Project:
            return self.Project.GetCustomBlockType(type, inputs)
        return None

    # Return Block types checking for recursion
    def GetBlockTypes(self, tagname = ""):
        type = None
        if self.Project:
            name = ""
            words = tagname.split("::")
            if words[0] in ["P","T","A"]:
                name = words[1]
                type = self.GetPouType(name)
        if type == "function":
            blocktypes = []
            for category in BlockTypes + self.PluginTypes:
                cat = {"name" : category["name"], "list" : []}
                for block in category["list"]:
                    if block["type"] == "function":
                        cat["list"].append(block)
                if len(cat["list"]) > 0:
                    blocktypes.append(cat)
        else:
            blocktypes = [category for category in BlockTypes + self.PluginTypes]
        if self.Project:
            blocktypes.append({"name" : "User-defined POUs", "list": self.Project.GetCustomBlockTypes(name)})
        return blocktypes

    # Return Function Block types checking for recursion
    def GetFunctionBlockTypes(self, tagname = ""):
        blocktypes = []
        for category in BlockTypes + self.PluginTypes:
            for block in category["list"]:
                if block["type"] != "function":
                    blocktypes.append(block["name"])
        if self.Project:
            name = ""
            words = tagname.split("::")
            if words[0] in ["P","T","A"]:
                name = words[1]
            blocktypes.extend(self.Project.GetCustomFunctionBlockTypes(name))
        return blocktypes

    # Return Block types checking for recursion
    def GetBlockResource(self):
        blocktypes = []
        for category in BlockTypes[:-1]:
            for blocktype in category["list"]:
                if blocktype["type"] == "program":
                    blocktypes.append(blocktype["name"])
        if self.Project:
            blocktypes.extend(self.Project.GetCustomBlockResource())
        return blocktypes

    # Return Data Types checking for recursion
    def GetDataTypes(self, tagname = "", basetypes = True):
        if basetypes:
            datatypes = self.GetBaseTypes()
        else:
            datatypes = []
        if self.Project:
            name = ""
            words = tagname.split("::")
            if words[0] in ["D"]:
                name = words[1]
            datatypes.extend(self.Project.GetCustomDataTypes(name))
        return datatypes

    # Return Base Type of given possible derived type
    def GetBaseType(self, type):
        if self.Project:
            return self.Project.GetBaseType(type)
        return None

    # Return Base Types
    def GetBaseTypes(self):
        return [value for value in TypeHierarchy.keys() if not value.startswith("ANY")]

    def IsOfType(self, type, reference):
        if self.Project:
            return self.Project.IsOfType(type, reference)
        elif reference is None:
            return True
        elif type == reference:
            return True
        else:
            if type in TypeHierarchy:
                return self.IsOfType(TypeHierarchy[type], reference)
        return None
    
    def IsEndType(self, type):
        if type is not None:
            return not type.startswith("ANY")
        return True

    def GetDataTypeRange(self, type):
        if self.Project:
            return self.Project.GetDataTypeRange(type)
        elif type in DataTypeRange:
            return DataTypeRange[type]
        return None
    
    # Return Subrange types
    def GetSubrangeBaseTypes(self, exclude):
        if self.Project:
            return self.Project.GetSubrangeBaseTypes(exclude)
        return []
    
    # Return Enumerated Values
    def GetEnumeratedDataValues(self):
        if self.Project:
            return self.Project.GetEnumeratedDataTypeValues()
        return []

#-------------------------------------------------------------------------------
#                   Project Element tag name computation functions
#-------------------------------------------------------------------------------
    
    # Compute a data type name
    def ComputeDataTypeName(self, datatype):
        return "D::%s" % datatype
    
    # Compute a pou name
    def ComputePouName(self, pou):
        return "P::%s" % pou
    
    # Compute a pou transition name
    def ComputePouTransitionName(self, pou, transition):
        return "T::%s::%s" % (pou, transition)
    
    # Compute a pou action name
    def ComputePouActionName(self, pou, action):
        return "A::%s::%s" % (pou, action)

    # Compute a pou  name
    def ComputeConfigurationName(self, config):
        return "C::%s" % config

    # Compute a pou  name
    def ComputeConfigurationResourceName(self, config, resource):
        return "R::%s::%s" % (config, resource)

    def GetElementType(self, tagname):
        words = tagname.split("::")
        return {"D" : ITEM_DATATYPE, "P" : ITEM_POU, 
                "T" : ITEM_TRANSITION, "A" : ITEM_ACTION,
                "C" : ITEM_CONFIGURATION, "R" : ITEM_RESOURCE}[words[0]]

#-------------------------------------------------------------------------------
#                    Project opened Data types management functions
#-------------------------------------------------------------------------------

    # Return the data type informations
    def GetDataTypeInfos(self, tagname):
        words = tagname.split("::")
        if words[0] == "D":
            infos = {}
            datatype = self.Project.getdataType(words[1])
            basetype_content = datatype.baseType.getcontent()
            if basetype_content["value"] is None:
                infos["type"] = "Directly"
                infos["base_type"] = basetype_content["name"]
            elif basetype_content["name"] == "derived":
                infos["type"] = "Directly"
                infos["base_type"] = basetype_content["value"].getname()
            elif basetype_content["name"] in ["subrangeSigned", "subrangeUnsigned"]:
                infos["type"] = "Subrange"
                infos["min"] = basetype_content["value"].range.getlower()
                infos["max"] = basetype_content["value"].range.getupper()
                base_type = basetype_content["value"].baseType.getcontent()
                if base_type["value"] is None:
                    infos["base_type"] = base_type["name"]
                else:
                    infos["base_type"] = base_type["value"].getname()
            elif basetype_content["name"] == "enum":
                infos["type"] = "Enumerated"
                infos["values"] = []
                for value in basetype_content["value"].values.getvalue():
                    infos["values"].append(value.getname())
            elif basetype_content["name"] == "array":
                infos["type"] = "Array"
                infos["dimensions"] = []
                for dimension in basetype_content["value"].getdimension():
                    infos["dimensions"].append((dimension.getlower(), dimension.getupper()))
                base_type = basetype_content["value"].baseType.getcontent()
                if base_type["value"] is None:
                    infos["base_type"] = base_type["name"]
                else:
                    infos["base_type"] = base_type["value"].getname()
            if datatype.initialValue is not None:
                infos["initial"] = str(datatype.initialValue.getvalue())
            else:
                infos["initial"] = ""
            return infos
        return None
    
    # Change the data type informations
    def SetDataTypeInfos(self, tagname, infos):
        words = tagname.split("::")
        if words[0] == "D":
            datatype = self.Project.getdataType(words[1])
            if infos["type"] == "Directly":
                if infos["base_type"] in self.GetBaseTypes():
                    if infos["base_type"] == "STRING":
                        datatype.baseType.setcontent({"name" : "string", "value" : plcopen.elementaryTypes_string()})
                    elif infos["base_type"] == "WSTRING":
                        datatype.baseType.setcontent({"name" : "wstring", "value" : plcopen.elementaryTypes_wstring()})
                    else:
                        datatype.baseType.setcontent({"name" : infos["base_type"], "value" : None})
                else:
                    derived_datatype = plcopen.derivedTypes_derived()
                    derived_datatype.setname(infos["base_type"])
                    datatype.baseType.setcontent({"name" : "derived", "value" : derived_datatype})
            elif infos["type"] == "Subrange":
                if infos["base_type"] in GetSubTypes("ANY_UINT"):
                    subrange = plcopen.derivedTypes_subrangeUnsigned()
                    datatype.baseType.setcontent({"name" : "subrangeUnsigned", "value" : subrange})
                else:
                    subrange = plcopen.derivedTypes_subrangeSigned()
                    datatype.baseType.setcontent({"name" : "subrangeSigned", "value" : subrange})
                subrange.range.setlower(infos["min"])
                subrange.range.setupper(infos["max"])
                if infos["base_type"] in self.GetBaseTypes():
                    subrange.baseType.setcontent({"name" : infos["base_type"], "value" : None})
                else:
                    derived_datatype = plcopen.derivedTypes_derived()
                    derived_datatype.setname(infos["base_type"])
                    subrange.baseType.setcontent({"name" : "derived", "value" : derived_datatype})
            elif infos["type"] == "Enumerated":
                enumerated = plcopen.derivedTypes_enum()
                for i, enum_value in enumerate(infos["values"]):
                    value = plcopen.values_value()
                    value.setname(enum_value)
                    if i == 0:
                        enumerated.values.setvalue([value])
                    else:
                        enumerated.values.appendvalue(value)
                datatype.baseType.setcontent({"name" : "enum", "value" : enumerated})
            elif infos["type"] == "Array":
                array = plcopen.derivedTypes_array()
                for i, dimension in enumerate(infos["dimensions"]):
                    dimension_range = plcopen.rangeSigned()
                    dimension_range.setlower(dimension[0])
                    dimension_range.setupper(dimension[1])
                    if i == 0:
                        array.setdimension([dimension_range])
                    else:
                        array.appenddimension(dimension_range)
                if infos["base_type"] in self.GetBaseTypes():
                    if infos["base_type"] == "STRING":
                        array.baseType.setcontent({"name" : "string", "value" : plcopen.elementaryTypes_string()})
                    elif infos["base_type"] == "WSTRING":
                        array.baseType.setcontent({"name" : "wstring", "value" : plcopen.wstring()})
                    else:
                        array.baseType.setcontent({"name" : infos["base_type"], "value" : None})
                else:
                    derived_datatype = plcopen.derivedTypes_derived()
                    derived_datatype.setname(infos["base_type"])
                    array.baseType.setcontent({"name" : "derived", "value" : derived_datatype})
                datatype.baseType.setcontent({"name" : "array", "value" : array})
            if infos["initial"] != "":
                if datatype.initialValue is None:
                    datatype.initialValue = plcopen.value()
                datatype.initialValue.setvalue(infos["initial"])
            else:
                datatype.initialValue = None
            self.Project.RefreshDataTypeHierarchy()
            self.BufferProject()
    
#-------------------------------------------------------------------------------
#                       Project opened Pous management functions
#-------------------------------------------------------------------------------

    # Return edited element
    def GetEditedElement(self, tagname):
        words = tagname.split("::")
        if words[0] == "D":
            return self.Project.getdataType(words[1])
        elif words[0] == "P":
            return self.Project.getpou(words[1])
        elif words[0] in ['T', 'A']:
            pou = self.Project.getpou(words[1])
            if words[0] == 'T':
                return pou.gettransition(words[2])
            elif words[0] == 'A':
                return pou.getaction(words[2])
        elif words[0] == 'C':
            return self.Project.getconfiguration(words[1])
        elif words[0] == 'R':
            return self.Project.getconfigurationResource(words[1], words[2])
        return None
    
    # Return edited element name
    def GetEditedElementName(self, tagname):
        words = tagname.split("::")
        if words[0] in ["P","C","D"]:
            return words[1]
        else:
            return words[2]
        return None
    
    # Return edited element name and type
    def GetEditedElementType(self, tagname):
        words = tagname.split("::")
        if words[0] in ["P","T","A"]:
            return words[1], self.GetPouType(words[1])
        return None, None

    # Return language in which edited element is written
    def GetEditedElementBodyType(self, tagname):
        words = tagname.split("::")
        if words[0] == "P":
            return self.GetPouBodyType(words[1])
        elif words[0] == 'T':
            return self.GetTransitionBodyType(words[1], words[2])
        elif words[0] == 'A':
            return self.GetActionBodyType(words[1], words[2])
        return None

    # Return the edited element variables
    def GetEditedElementInterfaceVars(self, tagname):
        words = tagname.split("::")
        if words[0] in ["P","T","A"]:
            pou = self.Project.getpou(words[1])
            return self.GetPouInterfaceVars(pou)
        return []

    # Return the edited element return type
    def GetEditedElementInterfaceReturnType(self, tagname):
        words = tagname.split("::")
        if words[0] == "P":
            pou = self.Project.getpou(words[1])
            return self.GetPouInterfaceReturnType(pou)
        elif words[0] == 'T':
            return "BOOL"
        return None
    
    # Change the edited element text
    def SetEditedElementText(self, tagname, text):
        element = self.GetEditedElement(tagname)
        if element != None:
            element.settext(text)
            self.Project.RefreshElementUsingTree()
    
    # Return the edited element text
    def GetEditedElementText(self, tagname):
        element = self.GetEditedElement(tagname)
        if element != None:
            return element.gettext()
        return ""

    # Return the edited element transitions
    def GetEditedElementTransitions(self, tagname):
        pou = self.GetEditedElement(tagname)
        if pou != None and pou.getbodyType() == "SFC":
            transitions = []
            for transition in pou.gettransitionList():
                transitions.append(transition.getname())
            return transitions
        return []

    # Return edited element transitions
    def GetEditedElementActions(self, tagname):
        pou = self.GetEditedElement(tagname)
        if pou != None and pou.getbodyType() == "SFC":
            actions = []
            for action in pou.getactionList():
                actions.append(action.getname())
            return actions
        return []

    # Return the names of the pou elements
    def GetEditedElementVariables(self, tagname):
        words = tagname.split("::")
        if words[0] in ["P","T","A"]:
            return self.GetProjectPouVariables(words[1])
        return []

    # Return the current pou editing informations
    def GetEditedElementInstanceInfos(self, tagname, id = None, exclude = []):
        infos = {}
        instance = None
        element = self.GetEditedElement(tagname)
        if element is not None:
            # if id is defined
            if id is not None:
                instance = element.getinstance(id)
            else:
                instance = element.getrandomInstance(exclude)
        if instance is not None:
            if id is not None:
                infos["id"] = id
            else:
                infos["id"] = instance.getlocalId() 
            infos["x"] = instance.getx()
            infos["y"] = instance.gety()
            infos["height"] = instance.getheight()
            infos["width"] = instance.getwidth()
            if isinstance(instance, plcopen.fbdObjects_block):
                infos["name"] = instance.getinstanceName()
                infos["type"] = instance.gettypeName()
                executionOrder = instance.getexecutionOrderId()
                if executionOrder is not None:
                    infos["executionOrder"] = executionOrder
                else:
                    infos["executionOrder"] = 0
                infos["connectors"] = {"inputs":[],"outputs":[]}
                for variable in instance.inputVariables.getvariable():
                    connector = {}
                    connector["name"] = variable.getformalParameter()
                    connector["position"] = variable.connectionPointIn.getrelPositionXY()
                    connector["negated"] = variable.getnegated()
                    connector["edge"] = variable.getedge()
                    connector["links"] = []
                    connections = variable.connectionPointIn.getconnections()
                    if connections:
                        for link in connections:
                            dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                            connector["links"].append(dic)
                    infos["connectors"]["inputs"].append(connector)
                for variable in instance.outputVariables.getvariable():
                    connector = {}
                    connector["name"] = variable.getformalParameter()
                    connector["position"] = variable.connectionPointOut.getrelPositionXY()
                    connector["negated"] = variable.getnegated()
                    connector["edge"] = variable.getedge()
                    infos["connectors"]["outputs"].append(connector)
            elif isinstance(instance, plcopen.fbdObjects_inVariable):
                infos["name"] = instance.getexpression()
                infos["value_type"] = self.GetEditedElementVarValueType(tagname, infos["name"])
                infos["type"] = "input"
                executionOrder = instance.getexecutionOrderId()
                if executionOrder is not None:
                    infos["executionOrder"] = executionOrder
                else:
                    infos["executionOrder"] = 0
                infos["connector"] = {}
                infos["connector"]["position"] = instance.connectionPointOut.getrelPositionXY()
                infos["connector"]["negated"] = instance.getnegated()
                infos["connector"]["edge"] = instance.getedge()
            elif isinstance(instance, plcopen.fbdObjects_outVariable):
                infos["name"] = instance.getexpression()
                infos["value_type"] = self.GetEditedElementVarValueType(tagname, infos["name"])
                infos["type"] = "output"
                executionOrder = instance.getexecutionOrderId()
                if executionOrder is not None:
                    infos["executionOrder"] = executionOrder
                else:
                    infos["executionOrder"] = 0
                infos["connector"] = {}
                infos["connector"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connector"]["negated"] = instance.getnegated()
                infos["connector"]["edge"] = instance.getedge()
                infos["connector"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connector"]["links"].append(dic)
            elif isinstance(instance, plcopen.fbdObjects_inOutVariable):
                infos["name"] = instance.getexpression()
                infos["value_type"] = self.GetEditedElementVarValueType(tagname, infos["name"])
                infos["type"] = "inout"
                executionOrder = instance.getexecutionOrderId()
                if executionOrder is not None:
                    infos["executionOrder"] = executionOrder
                else:
                    infos["executionOrder"] = 0
                infos["connectors"] = {"input":{},"output":{}}
                infos["connectors"]["output"]["position"] = instance.connectionPointOut.getrelPositionXY()
                infos["connectors"]["output"]["negated"] = instance.getnegatedOut()
                infos["connectors"]["output"]["edge"] = instance.getedgeOut()
                infos["connectors"]["input"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connectors"]["input"]["negated"] = instance.getnegatedIn()
                infos["connectors"]["input"]["edge"] = instance.getedgeIn()
                infos["connectors"]["input"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connectors"]["input"]["links"].append(dic)
            elif isinstance(instance, plcopen.commonObjects_continuation):
                infos["name"] = instance.getname()
                infos["type"] = "continuation"
                infos["connector"] = {}
                infos["connector"]["position"] = instance.connectionPointOut.getrelPositionXY()
            elif isinstance(instance, plcopen.commonObjects_connector):
                infos["name"] = instance.getname()
                infos["type"] = "connection"
                infos["connector"] = {}
                infos["connector"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connector"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connector"]["links"].append(dic)
            elif isinstance(instance, plcopen.commonObjects_comment):
                infos["type"] = "comment"
                infos["content"] = instance.getcontentText()
            elif isinstance(instance, plcopen.ldObjects_leftPowerRail):
                infos["type"] = "leftPowerRail"
                infos["connectors"] = []
                for connection in instance.getconnectionPointOut():
                    connector = {}
                    connector["position"] = connection.getrelPositionXY()
                    infos["connectors"].append(connector)
            elif isinstance(instance, plcopen.ldObjects_rightPowerRail):
                infos["type"] = "rightPowerRail"
                infos["connectors"] = []
                for connection in instance.getconnectionPointIn():
                    connector = {}
                    connector["position"] = connection.getrelPositionXY()
                    connector["links"] = []
                    connections = connection.getconnections()
                    if connections:
                        for link in connection.getconnections():
                            dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                            connector["links"].append(dic)
                    infos["connectors"].append(connector)
            elif isinstance(instance, plcopen.ldObjects_contact):
                infos["name"] = instance.getvariable()
                infos["type"] = "contact"
                executionOrder = instance.getexecutionOrderId()
                if executionOrder is not None:
                    infos["executionOrder"] = executionOrder
                else:
                    infos["executionOrder"] = 0
                infos["negated"] = instance.getnegated()
                infos["edge"] = instance.getedge()
                infos["connectors"] = {"input":{},"output":{}}
                infos["connectors"]["input"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connectors"]["input"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connectors"]["input"]["links"].append(dic)
                infos["connectors"]["output"]["position"] = instance.connectionPointOut.getrelPositionXY()
            elif isinstance(instance, plcopen.ldObjects_coil):
                infos["name"] = instance.getvariable()
                infos["type"] = "coil"
                executionOrder = instance.getexecutionOrderId()
                if executionOrder is not None:
                    infos["executionOrder"] = executionOrder
                else:
                    infos["executionOrder"] = 0
                infos["negated"] = instance.getnegated()
                infos["storage"] = instance.getstorage()
                infos["connectors"] = {"input":{},"output":{}}
                infos["connectors"]["input"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connectors"]["input"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connectors"]["input"]["links"].append(dic)
                infos["connectors"]["output"]["position"] = instance.connectionPointOut.getrelPositionXY()
            elif isinstance(instance, plcopen.sfcObjects_step):
                infos["name"] = instance.getname()
                infos["type"] = "step"
                infos["initial"] = instance.getinitialStep()
                infos["connectors"] = {}
                if instance.connectionPointIn:
                    infos["connectors"]["input"] = {}
                    infos["connectors"]["input"]["position"] = instance.connectionPointIn.getrelPositionXY()
                    infos["connectors"]["input"]["links"] = []
                    connections = instance.connectionPointIn.getconnections()
                    if connections:
                        for link in connections:
                            dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                            infos["connectors"]["input"]["links"].append(dic)
                if instance.connectionPointOut:
                    infos["connectors"]["output"] = {"position" : instance.connectionPointOut.getrelPositionXY()}
                if instance.connectionPointOutAction:
                    infos["connectors"]["action"] = {"position" : instance.connectionPointOutAction.getrelPositionXY()}
            elif isinstance(instance, plcopen.sfcObjects_transition):
                infos["type"] = "transition"
                condition = instance.getconditionContent()
                priority = instance.getpriority()
                if priority == None:
                    infos["priority"] = 0
                else:
                    infos["priority"] = priority
                infos["condition_type"] = condition["type"]
                infos["connectors"] = {"input":{},"output":{}}
                infos["connectors"]["input"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connectors"]["input"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connectors"]["input"]["links"].append(dic)
                infos["connectors"]["output"]["position"] = instance.connectionPointOut.getrelPositionXY()
                if infos["condition_type"] == "connection":
                    infos["connectors"]["connection"] = {}
                    infos["connectors"]["connection"]["links"] = []
                    connections = instance.getconnections()
                    if connections:
                        for link in connections:
                            dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                            infos["connectors"]["connection"]["links"].append(dic)
                    infos["condition"] = None
                else:
                    infos["condition"] = condition["value"]
            elif isinstance(instance, (plcopen.sfcObjects_selectionDivergence, plcopen.sfcObjects_simultaneousDivergence)):
                if isinstance(instance, plcopen.sfcObjects_selectionDivergence):
                    infos["type"] = "selectionDivergence"
                else:
                    infos["type"] = "simultaneousDivergence"
                infos["connectors"] = {"inputs":[],"outputs":[]}
                connector = {}
                connector["position"] = instance.connectionPointIn.getrelPositionXY()
                connector["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        connector["links"].append(dic)
                infos["connectors"]["inputs"].append(connector)
                for sequence in instance.getconnectionPointOut():
                    connector = {}
                    connector["position"] = sequence.getrelPositionXY()
                    infos["connectors"]["outputs"].append(connector)
            elif isinstance(instance, (plcopen.sfcObjects_selectionConvergence, plcopen.sfcObjects_simultaneousConvergence)):
                if isinstance(instance, plcopen.sfcObjects_selectionConvergence):
                    infos["type"] = "selectionConvergence"
                else:
                    infos["type"] = "simultaneousConvergence"
                infos["connectors"] = {"inputs":[],"outputs":[]}
                for sequence in instance.getconnectionPointIn():
                    connector = {}
                    connector["position"] = sequence.getrelPositionXY()
                    connector["links"] = []
                    connections = sequence.getconnections()
                    if connections:
                        for link in connections:
                            dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                            connector["links"].append(dic)
                    infos["connectors"]["inputs"].append(connector)
                connector = {}
                connector["position"] = instance.connectionPointOut.getrelPositionXY()
                infos["connectors"]["outputs"].append(connector)
            elif isinstance(instance, plcopen.sfcObjects_jumpStep):
                infos["type"] = "jump"
                infos["target"] = instance.gettargetName()
                infos["connector"] = {}
                infos["connector"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connector"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connector"]["links"].append(dic)
            elif isinstance(instance, plcopen.commonObjects_actionBlock):
                infos["type"] = "actionBlock"
                infos["actions"] = instance.getactions()
                infos["connector"] = {}
                infos["connector"]["position"] = instance.connectionPointIn.getrelPositionXY()
                infos["connector"]["links"] = []
                connections = instance.connectionPointIn.getconnections()
                if connections:
                    for link in connections:
                        dic = {"refLocalId":link.getrefLocalId(),"points":link.getpoints(),"formalParameter":link.getformalParameter()}
                        infos["connector"]["links"].append(dic)
            return infos
        return False
    
    def ClearEditedElementExecutionOrder(self, tagname):
        element = self.GetEditedElement(tagname)
        if element is not None:
            element.resetexecutionOrder()
    
    def ResetEditedElementExecutionOrder(self, tagname):
        element = self.GetEditedElement(tagname)
        if element is not None:
            element.compileexecutionOrder()
    
    # Return the variable type of the given pou
    def GetEditedElementVarValueType(self, tagname, varname):
        words = tagname.split("::")
        if words[0] in ["P","T","A"]:
            pou = self.Project.getpou(words[1])
            for type, varlist in pou.getvars():
                for var in varlist.getvariable():
                    if var.getname() == varname:
                        vartype_content = var.gettype().getcontent()
                        if vartype_content["name"] == "derived":
                            return vartype_content["value"].getname()
                        elif vartype_content["name"] in ["string", "wstring"]:
                            return vartype_content["name"].upper()
                        else:
                            return vartype_content["name"]
        return None
    
    def SetConnectionWires(self, connection, connector):
        wires = connector.GetWires()
        idx = 0
        for wire, handle in wires:
            points = wire.GetPoints(handle != 0)
            if handle == 0:
                result = wire.GetConnectedInfos(-1)
            else:
                result = wire.GetConnectedInfos(0)
            if result != None:
                refLocalId, formalParameter = result
                connections = connection.getconnections()
                if connections is None or len(connection.getconnections()) <= idx:
                    connection.addconnection()
                connection.setconnectionId(idx, refLocalId)
                connection.setconnectionPoints(idx, points)
                if formalParameter != "":
                    connection.setconnectionParameter(idx, formalParameter)
                else:
                    connection.setconnectionParameter(idx, None)
                idx += 1
    
    def AddEditedElementPouVar(self, tagname, type, name):
        words = tagname.split("::")
        if words[0] in ['P', 'T', 'A']:
            pou = self.Project.getpou(words[1])
            if not pou.interface:
                pou.interface = plcopen.pou_interface()
            pou.addpouVar(type, name)
            
    def ChangeEditedElementPouVar(self, tagname, old_type, old_name, new_type, new_name):
        words = tagname.split("::")
        if words[0] in ['P', 'T', 'A']:
            pou = self.Project.getpou(words[1])
            pou.changepouVar(old_type, old_name, new_type, new_name)
    
    def RemoveEditedElementPouVar(self, tagname, type, name):
        words = tagname.split("::")
        if words[0] in ['P', 'T', 'A']:
            pou = self.Project.getpou(words[1])
            pou.removepouVar(type, name)
    
    def AddEditedElementBlock(self, tagname, id, blocktype, blockname = None):
        element = self.GetEditedElement(tagname)
        if element is not None:
            block = plcopen.fbdObjects_block()
            block.setlocalId(id)
            block.settypeName(blocktype)
            blocktype_infos = self.GetBlockType(blocktype)
            if blocktype_infos["type"] != "function" and blockname is not None:
                block.setinstanceName(blockname)
                self.AddEditedElementPouVar(tagname, blocktype, blockname)
            element.addinstance("block", block)
            self.Project.RefreshElementUsingTree()
    
    def SetEditedElementBlockInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            block = element.getinstance(id)
            if block is None:
                return 
            old_name = block.getinstanceName()
            old_type = block.gettypeName()
            new_name = infos.get("name", old_name)
            new_type = infos.get("type", old_type)
            old_typeinfos = self.GetBlockType(old_type)
            new_typeinfos = self.GetBlockType(new_type)
            if new_typeinfos["type"] != old_typeinfos["type"]:
                if new_typeinfos["type"] == "function":
                    self.RemoveEditedElementPouVar(tagname, old_type, old_name)
                else:
                    self.AddEditedElementPouVar(tagname, new_type, new_name)
            elif new_typeinfos["type"] != "function" and old_name != new_name:
                self.ChangeEditedElementPouVar(tagname, old_type, old_name, new_type, new_name)
            for param, value in infos.items():
                if param == "name":
                    block.setinstanceName(value)
                elif param == "type":
                    block.settypeName(value)
                elif param == "executionOrder" and block.getexecutionOrderId() != value:
                    element.setelementExecutionOrder(block, value)
                elif param == "height":
                    block.setheight(value)
                elif param == "width":
                    block.setwidth(value)
                elif param == "x":
                    block.setx(value)
                elif param == "y":
                    block.sety(value)
                elif param == "connectors":
                    block.inputVariables.setvariable([])
                    block.outputVariables.setvariable([])
                    for connector in value["inputs"]:
                        variable = plcopen.inputVariables_variable()
                        variable.setformalParameter(connector.GetName())
                        if connector.IsNegated():
                            variable.setnegated(True)
                        if connector.GetEdge() != "none":
                            variable.setedge(connector.GetEdge())
                        position = connector.GetRelPosition()
                        variable.connectionPointIn.setrelPositionXY(position.x, position.y)
                        self.SetConnectionWires(variable.connectionPointIn, connector)
                        block.inputVariables.appendvariable(variable)
                    for connector in value["outputs"]:
                        variable = plcopen.outputVariables_variable()
                        variable.setformalParameter(connector.GetName())
                        if connector.IsNegated():
                            variable.setnegated(True)
                        if connector.GetEdge() != "none":
                            variable.setedge(connector.GetEdge())
                        position = connector.GetRelPosition()
                        variable.addconnectionPointOut()
                        variable.connectionPointOut.setrelPositionXY(position.x, position.y)
                        block.outputVariables.appendvariable(variable)
            self.Project.RefreshElementUsingTree()
        
    def AddEditedElementVariable(self, tagname, id, type):
        element = self.GetEditedElement(tagname)
        if element is not None:            
            if type == INPUT:
                name = "inVariable"
                variable = plcopen.fbdObjects_inVariable()
            elif type == OUTPUT:
                name = "outVariable"
                variable = plcopen.fbdObjects_outVariable()
            elif type == INOUT:
                name = "inOutVariable"
                variable = plcopen.fbdObjects_inOutVariable()
            variable.setlocalId(id)
            element.addinstance(name, variable)
        
    def SetEditedElementVariableInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            variable = element.getinstance(id)
            if variable is None:
                return 
            for param, value in infos.items():
                if param == "name":
                    variable.setexpression(value)    
                elif param == "executionOrder" and variable.getexecutionOrderId() != value:
                    element.setelementExecutionOrder(variable, value)
                elif param == "height":
                    variable.setheight(value)
                elif param == "width":
                    variable.setwidth(value)
                elif param == "x":
                    variable.setx(value)
                elif param == "y":
                    variable.sety(value)
                elif param == "connectors":
                    if isinstance(variable, plcopen.fbdObjects_inVariable):
                        if value["output"].IsNegated():
                            variable.setnegated(True)
                        if value["output"].GetEdge() != "none":
                            variable.setedge(value["output"].GetEdge())
                        position = value["output"].GetRelPosition()
                        variable.addconnectionPointOut()
                        variable.connectionPointOut.setrelPositionXY(position.x, position.y)
                    elif isinstance(variable, plcopen.fbdObjects_outVariable):
                        if value["input"].IsNegated():
                            variable.setnegated(True)
                        if value["input"].GetEdge() != "none":
                            variable.setedge(value["input"].GetEdge())
                        position = value["input"].GetRelPosition()
                        variable.addconnectionPointIn()
                        variable.connectionPointIn.setrelPositionXY(position.x, position.y)
                        self.SetConnectionWires(variable.connectionPointIn, value["input"])
                    elif isinstance(variable, plcopen.fbdObjects_inOutVariable):
                        if value["input"].IsNegated():
                            variable.setnegatedIn(True)
                        if value["input"].GetEdge() != "none":
                            variable.setedgeIn(value["input"].GetEdge())
                        if value["output"].IsNegated():
                            variable.setnegatedOut(True)
                        if value["output"].GetEdge() != "none":
                            variable.setedgeOut(value["output"].GetEdge())
                        position = value["output"].GetRelPosition()
                        variable.addconnectionPointOut()
                        variable.connectionPointOut.setrelPositionXY(position.x, position.y)
                        position = value["input"].GetRelPosition()
                        variable.addconnectionPointIn()
                        variable.connectionPointIn.setrelPositionXY(position.x, position.y)
                        self.SetConnectionWires(variable.connectionPointIn, value["input"])

    def AddEditedElementConnection(self, tagname, id, type):
        element = self.GetEditedElement(tagname)
        if element is not None:
            if type == CONNECTOR:
                name = "connector"
                connection = plcopen.commonObjects_connector()
            elif type == CONTINUATION:
                name = "continuation"
                connection = plcopen.commonObjects_continuation()
            connection.setlocalId(id)
            element.addinstance(name, connection)
        
    def SetEditedElementConnectionInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            connection = element.getinstance(id)
            if connection is None:
                return
            for param, value in infos.items():
                if param == "name":
                    connection.setname(value)    
                elif param == "height":
                    connection.setheight(value)
                elif param == "width":
                    connection.setwidth(value)
                elif param == "x":
                    connection.setx(value)
                elif param == "y":
                    connection.sety(value)
                elif param == "connector":
                    position = value.GetRelPosition()
                    if isinstance(connection, plcopen.commonObjects_continuation):
                        connection.addconnectionPointOut()
                        connection.connectionPointOut.setrelPositionXY(position.x, position.y)
                    elif isinstance(connection, plcopen.commonObjects_connector):
                        connection.addconnectionPointIn()
                        connection.connectionPointIn.setrelPositionXY(position.x, position.y)
                        self.SetConnectionWires(connection.connectionPointIn, value)

    def AddEditedElementComment(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            comment = plcopen.commonObjects_comment()
            comment.setlocalId(id)
            element.addinstance("comment", comment)
    
    def SetEditedElementCommentInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            comment = element.getinstance(id)
            for param, value in infos.items():
                if param == "content":
                    comment.setcontentText(value)
                elif param == "height":
                    comment.setheight(value)
                elif param == "width":
                    comment.setwidth(value)
                elif param == "x":
                    comment.setx(value)
                elif param == "y":
                    comment.sety(value)

    def AddEditedElementPowerRail(self, tagname, id, type):
        element = self.GetEditedElement(tagname)
        if element is not None:
            if type == LEFTRAIL:
                name = "leftPowerRail"
                powerrail = plcopen.ldObjects_leftPowerRail()
            elif type == RIGHTRAIL:
                name = "rightPowerRail"
                powerrail = plcopen.ldObjects_rightPowerRail()
            powerrail.setlocalId(id)
            element.addinstance(name, powerrail)
    
    def SetEditedElementPowerRailInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            powerrail = element.getinstance(id)
            if powerrail is None:
                return
            for param, value in infos.items():
                if param == "height":
                    powerrail.setheight(value)
                elif param == "width":
                    powerrail.setwidth(value)
                elif param == "x":
                    powerrail.setx(value)
                elif param == "y":
                    powerrail.sety(value)
                elif param == "connectors":
                    if isinstance(powerrail, plcopen.ldObjects_leftPowerRail):
                        powerrail.setconnectionPointOut([])
                        for connector in value:
                            position = connector.GetRelPosition()
                            connection = plcopen.leftPowerRail_connectionPointOut()
                            connection.setrelPositionXY(position.x, position.y)
                            powerrail.connectionPointOut.append(connection)
                    elif isinstance(powerrail, plcopen.ldObjects_rightPowerRail):
                        powerrail.setconnectionPointIn([])
                        for connector in value:
                            position = connector.GetRelPosition()
                            connection = plcopen.connectionPointIn()
                            connection.setrelPositionXY(position.x, position.y)
                            self.SetConnectionWires(connection, connector)
                            powerrail.connectionPointIn.append(connection)

    def AddEditedElementContact(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            contact = plcopen.ldObjects_contact()
            contact.setlocalId(id)
            element.addinstance("contact", contact)

    def SetEditedElementContactInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            contact = element.getinstance(id)
            if contact is None:
                return
            for param, value in infos.items():
                if param == "name":
                    contact.setvariable(value)
                elif param == "type":
                    if value == CONTACT_NORMAL:
                        contact.setnegated(False)
                        contact.setedge("none")
                    elif value == CONTACT_REVERSE:
                        contact.setnegated(True)
                        contact.setedge("none")
                    elif value == CONTACT_RISING:
                        contact.setnegated(False)
                        contact.setedge("rising")
                    elif value == CONTACT_FALLING:
                        contact.setnegated(False)
                        contact.setedge("falling")
                elif param == "height":
                    contact.setheight(value)
                elif param == "width":
                    contact.setwidth(value)
                elif param == "x":
                    contact.setx(value)
                elif param == "y":
                    contact.sety(value)
                elif param == "connectors":
                    input_connector = value["input"]
                    position = input_connector.GetRelPosition()
                    contact.addconnectionPointIn()
                    contact.connectionPointIn.setrelPositionXY(position.x, position.y)
                    self.SetConnectionWires(contact.connectionPointIn, input_connector)
                    output_connector = value["output"]
                    position = output_connector.GetRelPosition()
                    contact.addconnectionPointOut()
                    contact.connectionPointOut.setrelPositionXY(position.x, position.y)

    def AddEditedElementCoil(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            coil = plcopen.ldObjects_coil()
            coil.setlocalId(id)
            element.addinstance("coil", coil)

    def SetEditedElementCoilInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            coil = element.getinstance(id)
            if coil is None:
                return
            for param, value in infos.items():
                if param == "name":
                    coil.setvariable(value)
                elif param == "type":
                    if value == COIL_NORMAL:
                        coil.setnegated(False)
                        coil.setstorage("none")
                    elif value == COIL_REVERSE:
                        coil.setnegated(True)
                        coil.setstorage("none")
                    elif value == COIL_SET:
                        coil.setnegated(False)
                        coil.setstorage("set")
                    elif value == COIL_RESET:
                        coil.setnegated(False)
                        coil.setstorage("reset")
                elif param == "height":
                    coil.setheight(value)
                elif param == "width":
                    coil.setwidth(value)
                elif param == "x":
                    coil.setx(value)
                elif param == "y":
                    coil.sety(value)
                elif param == "connectors":
                    input_connector = value["input"]
                    position = input_connector.GetRelPosition()
                    coil.addconnectionPointIn()
                    coil.connectionPointIn.setrelPositionXY(position.x, position.y)
                    self.SetConnectionWires(coil.connectionPointIn, input_connector)
                    output_connector = value["output"]
                    position = output_connector.GetRelPosition()
                    coil.addconnectionPointOut()
                    coil.connectionPointOut.setrelPositionXY(position.x, position.y)

    def AddEditedElementStep(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            step = plcopen.sfcObjects_step()
            step.setlocalId(id)
            element.addinstance("step", step)
    
    def SetEditedElementStepInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            step = element.getinstance(id)
            if step is None:
                return
            for param, value in infos.items():
                if param == "name":
                    step.setname(value)
                elif param == "initial":
                    step.setinitialStep(value)
                elif param == "height":
                    step.setheight(value)
                elif param == "width":
                    step.setwidth(value)
                elif param == "x":
                    step.setx(value)
                elif param == "y":
                    step.sety(value)
                elif param == "connectors":
                    input_connector = value["input"]
                    if input_connector:
                        position = input_connector.GetRelPosition()
                        step.addconnectionPointIn()
                        step.connectionPointIn.setrelPositionXY(position.x, position.y)
                        self.SetConnectionWires(step.connectionPointIn, input_connector)
                    else:
                        step.deleteconnectionPointIn()
                    output_connector = value["output"]
                    if output_connector:
                        position = output_connector.GetRelPosition()
                        step.addconnectionPointOut()
                        step.connectionPointOut.setrelPositionXY(position.x, position.y)
                    else:
                        step.deleteconnectionPointOut()
                    action_connector = value["action"]
                    if action_connector:
                        position = action_connector.GetRelPosition()
                        step.addconnectionPointOutAction()
                        step.connectionPointOutAction.setrelPositionXY(position.x, position.y)
                    else:
                        step.deleteconnectionPointOutAction()
    
    def AddEditedElementTransition(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            transition = plcopen.sfcObjects_transition()
            transition.setlocalId(id)
            element.addinstance("transition", transition)
    
    def SetEditedElementTransitionInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            transition = element.getinstance(id)
            if transition is None:
                return
            for param, value in infos.items():
                if param == "type" and value != "connection":
                    transition.setconditionContent(value, infos["condition"])
                elif param == "height":
                    transition.setheight(value)
                elif param == "width":
                    transition.setwidth(value)
                elif param == "x":
                    transition.setx(value)
                elif param == "y":
                    transition.sety(value)
                elif param == "priority":
                    if value != 0:
                        transition.setpriority(value)
                    else:
                        transition.setpriority(None)
                elif param == "connectors":
                    input_connector = value["input"]
                    position = input_connector.GetRelPosition()
                    transition.addconnectionPointIn()
                    transition.connectionPointIn.setrelPositionXY(position.x, position.y)
                    self.SetConnectionWires(transition.connectionPointIn, input_connector)
                    output_connector = value["output"]
                    position = output_connector.GetRelPosition()
                    transition.addconnectionPointOut()
                    transition.connectionPointOut.setrelPositionXY(position.x, position.y)
                    if infos.get("type", None) == "connection":
                        transition.setconditionContent("connection", None)
                        connection_connector = value["connection"]
                        self.SetConnectionWires(transition, connection_connector)
    
    def AddEditedElementDivergence(self, tagname, id, type):
        element = self.GetEditedElement(tagname)
        if element is not None:
            if type == SELECTION_DIVERGENCE:
                name = "selectionDivergence"
                divergence = plcopen.sfcObjects_selectionDivergence()
            elif type == SELECTION_CONVERGENCE:
                name = "selectionConvergence"
                divergence = plcopen.sfcObjects_selectionConvergence()
            elif type == SIMULTANEOUS_DIVERGENCE:
                name = "simultaneousDivergence"
                divergence = plcopen.sfcObjects_simultaneousDivergence()
            elif type == SIMULTANEOUS_CONVERGENCE:
                name = "simultaneousConvergence"
                divergence = plcopen.sfcObjects_simultaneousConvergence()
            divergence.setlocalId(id)
            element.addinstance(name, divergence)
    
    def SetEditedElementDivergenceInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            divergence = element.getinstance(id)
            if divergence is None:
                return
            for param, value in infos.items():
                if param == "height":
                    divergence.setheight(value)
                elif param == "width":
                    divergence.setwidth(value)
                elif param == "x":
                    divergence.setx(value)
                elif param == "y":
                    divergence.sety(value)
                elif param == "connectors":
                    input_connectors = value["inputs"]
                    if isinstance(divergence, (plcopen.sfcObjects_selectionDivergence, plcopen.sfcObjects_simultaneousDivergence)):
                        position = input_connectors[0].GetRelPosition()
                        divergence.addconnectionPointIn()
                        divergence.connectionPointIn.setrelPositionXY(position.x, position.y)
                        self.SetConnectionWires(divergence.connectionPointIn, input_connectors[0])
                    else:
                        divergence.setconnectionPointIn([])
                        for input_connector in input_connectors:
                            position = input_connector.GetRelPosition()
                            if isinstance(divergence, plcopen.sfcObjects_selectionConvergence):
                                connection = plcopen.selectionConvergence_connectionPointIn()
                            else:
                                connection = plcopen.connectionPointIn()
                            connection.setrelPositionXY(position.x, position.y)
                            self.SetConnectionWires(connection, input_connector)
                            divergence.appendconnectionPointIn(connection)
                    output_connectors = value["outputs"]
                    if isinstance(divergence, (plcopen.sfcObjects_selectionConvergence, plcopen.sfcObjects_simultaneousConvergence)):
                        position = output_connectors[0].GetRelPosition()
                        divergence.addconnectionPointOut()
                        divergence.connectionPointOut.setrelPositionXY(position.x, position.y)
                    else:
                        divergence.setconnectionPointOut([])
                        for output_connector in output_connectors:
                            position = output_connector.GetRelPosition()
                            if isinstance(divergence, plcopen.sfcObjects_selectionDivergence):
                                connection = plcopen.selectionDivergence_connectionPointOut()
                            else:
                                connection = plcopen.simultaneousDivergence_connectionPointOut()
                            connection.setrelPositionXY(position.x, position.y)
                            divergence.appendconnectionPointOut(connection)
    
    def AddEditedElementJump(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            jump = plcopen.sfcObjects_jumpStep()
            jump.setlocalId(id)
            element.addinstance("jumpStep", jump)
    
    def SetEditedElementJumpInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            jump = element.getinstance(id)
            if jump is None:
                return
            for param, value in infos.items():
                if param == "target":
                    jump.settargetName(value)
                elif param == "height":
                    jump.setheight(value)
                elif param == "width":
                    jump.setwidth(value)
                elif param == "x":
                    jump.setx(value)
                elif param == "y":
                    jump.sety(value)
                elif param == "connector":
                    position = value.GetRelPosition()
                    jump.addconnectionPointIn()
                    jump.connectionPointIn.setrelPositionXY(position.x, position.y)
                    self.SetConnectionWires(jump.connectionPointIn, value)
 
    def AddEditedElementActionBlock(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            actionBlock = plcopen.commonObjects_actionBlock()
            actionBlock.setlocalId(id)
            element.addinstance("actionBlock", actionBlock)
    
    def SetEditedElementActionBlockInfos(self, tagname, id, infos):
        element = self.GetEditedElement(tagname)
        if element is not None:
            actionBlock = element.getinstance(id)
            if actionBlock is None:
                return
            for param, value in infos.items():
                if param == "actions":
                    actionBlock.setactions(value)
                elif param == "height":
                    actionBlock.setheight(value)
                elif param == "width":
                    actionBlock.setwidth(value)
                elif param == "x":
                    actionBlock.setx(value)
                elif param == "y":
                    actionBlock.sety(value)
                elif param == "connector":
                    position = value.GetRelPosition()
                    actionBlock.addconnectionPointIn()
                    actionBlock.connectionPointIn.setrelPositionXY(position.x, position.y)
                    self.SetConnectionWires(actionBlock.connectionPointIn, value)
    
    def RemoveEditedElementInstance(self, tagname, id):
        element = self.GetEditedElement(tagname)
        if element is not None:
            instance = element.getinstance(id)
            if isinstance(instance, plcopen.fbdObjects_block):
                self.RemoveEditedElementPouVar(tagname, instance.gettypeName(), instance.getinstanceName())
            element.removeinstance(id)
            self.Project.RefreshElementUsingTree()

    def GetEditedResourceVariables(self, tagname):
        varlist = []
        words = tagname.split("::")
        for var in self.GetConfigurationGlobalVars(words[1]):
            if var["Type"] == "BOOL":
                varlist.append(var["Name"])
        for var in self.GetConfigurationResourceGlobalVars(words[1], words[2]):
            if var["Type"] == "BOOL":
                varlist.append(var["Name"])
        return varlist

    def SetEditedResourceInfos(self, tagname, tasks, instances):
        resource = self.GetEditedElement(tagname)
        if resource is not None:
            resource.settask([])
            resource.setpouInstance([])
            task_list = {}
            for task in tasks:
                new_task = plcopen.resource_task()
                new_task.setname(task["Name"])
                if task["Single"] != "":
                    new_task.setsingle(task["Single"])
                result = duration_model.match(task["Interval"]).groups()
                if reduce(lambda x, y: x or y != None, result):
                    values = []
                    for value in result[:-1]:
                        if value != None:
                            values.append(int(value))
                        else:
                            values.append(0)
                    if result[-1] is not None:
                        values.append(int(float(result[-1]) * 1000))
                    new_task.setinterval(time(*values))
                new_task.setpriority(int(task["Priority"]))
                if task["Name"] != "":
                    task_list[task["Name"]] = new_task
                resource.appendtask(new_task)
            for instance in instances:
                new_instance = plcopen.pouInstance()
                new_instance.setname(instance["Name"])
                new_instance.settype(instance["Type"])
                if instance["Task"] != "":
                    task_list[instance["Task"]].appendpouInstance(new_instance)
                else:
                    resource.appendpouInstance(new_instance)

    def GetEditedResourceInfos(self, tagname):
        resource = self.GetEditedElement(tagname)
        if resource is not None:
            tasks = resource.gettask()
            instances = resource.getpouInstance()
            tasks_data = []
            instances_data = []
            for task in tasks:
                new_task = {}
                new_task["Name"] = task.getname()
                single = task.getsingle()
                if single:
                    new_task["Single"] = single
                else:
                    new_task["Single"] = ""
                interval = task.getinterval()
                if interval:
                    text = ""
                    if interval.hour != 0:
                        text += "%dh"%interval.hour
                    if interval.minute != 0:
                        text += "%dm"%interval.minute
                    if interval.second != 0:
                        text += "%ds"%interval.second
                    if interval.microsecond != 0:
                        if interval.microsecond % 1000 != 0:
                            text += "%.3fms"%(float(interval.microsecond) / 1000)
                        else:
                            text += "%dms"%(interval.microsecond / 1000)
                    new_task["Interval"] = text
                else:
                    new_task["Interval"] = ""
                new_task["Priority"] = str(task.getpriority())
                tasks_data.append(new_task)
                for instance in task.getpouInstance():
                    new_instance = {}
                    new_instance["Name"] = instance.getname()
                    new_instance["Type"] = instance.gettype()
                    new_instance["Task"] = task.getname()
                    instances_data.append(new_instance)
            for instance in instances:
                new_instance = {}
                new_instance["Name"] = instance.getname()
                new_instance["Type"] = instance.gettype()
                new_instance["Task"] = ""
                instances_data.append(new_instance)
            return tasks_data, instances_data

    def OpenXMLFile(self, filepath):
        xmlfile = open(filepath, 'r')
        tree = minidom.parse(xmlfile)
        xmlfile.close()
        
        self.Project = plcopen.project()
        for child in tree.childNodes:
            if child.nodeType == tree.ELEMENT_NODE and child.nodeName == "project":
                self.Project.loadXMLTree(child, ["xmlns", "xmlns:xhtml", "xmlns:xsi", "xsi:schemaLocation"])
                self.SetFilePath(filepath)
                self.Project.RefreshElementUsingTree()
                self.Project.RefreshDataTypeHierarchy()
                self.Project.RefreshCustomBlockTypes()
                self.ProjectBuffer = UndoBuffer(self.Copy(self.Project), True)
                self.Buffering = False
                self.ElementsOpened = []
                self.CurrentElementEditing = None
                return None
        return "No PLC project found"

    def SaveXMLFile(self, filepath = None):
        if not filepath and self.FilePath == "":
            return False
        else:
            contentheader = self.Project.getcontentHeader()
            contentheader["modificationDateTime"] = datetime(*localtime()[:6])
            self.Project.setcontentHeader(contentheader)
            
            text = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
            extras = {"xmlns" : "http://www.plcopen.org/xml/tc6.xsd",
                      "xmlns:xhtml" : "http://www.w3.org/1999/xhtml",
                      "xmlns:xsi" : "http://www.w3.org/2001/XMLSchema-instance",
                      "xsi:schemaLocation" : "http://www.plcopen.org/xml/tc6.xsd http://www.plcopen.org/xml/tc6.xsd"}
            text += self.Project.generateXMLText("project", 0, extras)
            
            if filepath:
                xmlfile = open(filepath,"w")
            else:
                xmlfile = open(self.FilePath,"w")
            xmlfile.write(text)
            xmlfile.close()
            self.ProjectBuffer.CurrentSaved()
            if filepath:
                self.SetFilePath(filepath)
            return True

#-------------------------------------------------------------------------------
#                      Current Buffering Management Functions
#-------------------------------------------------------------------------------

    """
    Return a copy of the project
    """
    def Copy(self, model):
        return cPickle.loads(cPickle.dumps(model))

    def BufferProject(self):
        self.ProjectBuffer.Buffering(self.Copy(self.Project))

    def StartBuffering(self):
        self.ProjectBuffer.Buffering(self.Project)
        self.Buffering = True
        
    def EndBuffering(self):
        if self.Buffering:
            self.Project = self.Copy(self.Project)
            self.Buffering = False

    def ProjectIsSaved(self):
        if self.ProjectBuffer:
            return self.ProjectBuffer.IsCurrentSaved()
        else:
            return True

    def LoadPrevious(self):
        self.Project = self.Copy(self.ProjectBuffer.Previous())
    
    def LoadNext(self):
        self.Project = self.Copy(self.ProjectBuffer.Next())
    
    def GetBufferState(self):
        first = self.ProjectBuffer.IsFirst()
        last = self.ProjectBuffer.IsLast()
        return not first, not last
